@@ -171,7 +171,10 @@ let filter_unnec_from_propmatrix pcore =
       )
     !clauses
 
+let propmatrixh : (int,literal list) Hashtbl.t = Hashtbl.create 100;;
+
 let printf_fof_file_propmatrix s =
+  Hashtbl.clear propmatrixh;
   let axc = ref 0 in
   List.iter (fun c ->
     match c with
@@ -179,6 +182,7 @@ let printf_fof_file_propmatrix s =
 	if forced_into_propmatrix c || not (Hashtbl.mem omit_from_propmatrix c) then
 	  begin
 	    incr axc;
+	    Hashtbl.add propmatrixh !axc (l1::ll);
 	    Printf.fprintf s "fof(ax%d,axiom,( %s " (!axc) (plit_fo_str l1);
 	    List.iter (fun l2 -> Printf.fprintf s " | %s" (plit_fo_str l2)) ll;
 	    Printf.fprintf s ")).\n"
@@ -814,6 +818,11 @@ let call_e b =
 		| None ->
 		    ignore(Unix.close_process_full(myout,myin,myerr));
 		    raise (Unsatisfiable None)
+		| Some(TSTP) ->
+                    let (pcore,fcore) = read_e_core myout in
+		    ignore(Unix.close_process_full(myout,myin,myerr));
+		    let cri = !clause_ruleinfo in
+		    raise (Unsatisfiable(Some(Eprover(List.map (fun z -> Hashtbl.find propmatrixh z) pcore,Hashtbl.find cri,fcore))))
 		| Some(ptk) ->
 		    enable_pattern_clauses := false; (*** since we aren't doing general search anymore, turn off pattern clauses ***)
                     let ecore = read_e_core myout in
