@@ -1,4 +1,4 @@
-open State
+open Refut
 open Syntax
 open Branch
 open Litcount
