@@ -12,13 +12,13 @@ open Minisat
 
 exception CoqProofTooBig of int
 
-
 let clauses : clause list ref = ref []
 let clausesTable : (clause,unit) Hashtbl.t = Hashtbl.create 10
 
 let slaveargs = ref [Sys.argv.(0)];;
 let mode : string list ref = ref []
 let timeout : float option ref = ref None
+let hardtimeout : float option ref = ref None
 let nontheorem : bool ref = ref false (*** March 2012 - Know if you're explicitly trying to determine Satisfiable/CounterSatisfiable ***)
 let coq = ref false
 let coq2 = ref false
@@ -31,9 +31,15 @@ let coqinticks : ((out_channel -> unit) option * int) list ref = ref []
 let coqoutfn1 = ref (fun c -> ())
 let coqctx : (string option * pretrm option * pretrm option) list ref = ref []
 let coqglobalsectionstack : (string * (out_channel -> unit) * (string option * pretrm option * pretrm option) list) list ref = ref []
+let recognize_hashroots = ref false
 
 let training_file : string option ref = ref None
 
+type probitem =
+  | ProbDef of string * stp * trm * (string * string) list * float
+  | ProbAx of string * string * trm * (string * string) list * float
+  | ProbConj of string * trm * (string * string) list * float
+let probsig : probitem list ref = ref []
 
 let rec updatecoqglobalsectionstack cx cgss co =
   match cgss with
@@ -41,25 +47,32 @@ let rec updatecoqglobalsectionstack cx cgss co =
   | [] -> []
 
 let conjecturename : string ref = ref "claim"
-let conjecture : (pretrm * trm * trm) option ref = ref None
-type proofkind = TSTP | CoqScript | CoqSPfTerm | HOCore | Model | ModelTrue | IsarScript
+let conjecture : (trm * trm) option ref = ref None
+type proofkind = TSTP | CoqScript | CoqSPfTerm | HOCore | Model | ModelTrue | IsarScript | PfInfo | PfUseful | PfFormdeps
 let mkproofterm = ref None
 let mkprooftermp () = Utils.bool_of_option !mkproofterm || Utils.bool_of_option !training_file
+let pfusefulout = ref None
+let pfformdepsout = ref None
 let slave = ref false
 let coqsig_base : string list ref = ref []
 let coqsig_const : (string * stp) list ref = ref []
 let coqsig_def : (string * pretrm) list ref = ref []
-let coqsig_hyp : (string * pretrm) list ref = ref []
 let coqsig_def_trm : (string * trm) list ref = ref []
+let coqsig_hyp : (string * pretrm) list ref = ref []
 let coqsig_hyp_trm : (string * trm) list ref = ref []
 let name_base : (string,unit) Hashtbl.t = Hashtbl.create 10
 let name_base_list : string list ref = ref []
 let name_tp : (string,stp) Hashtbl.t = Hashtbl.create 100
 let name_trm : (string,(trm * stp) * bool ref) Hashtbl.t = Hashtbl.create 100
 let name_trm_list : (string * trm * stp) list ref = ref []
+let translucent_defns : bool ref = ref false
+let name_def_empty : (string,trm) Hashtbl.t = Hashtbl.create 1
 let name_def : (string,trm) Hashtbl.t = Hashtbl.create 100
+let name_def_all : (string,trm) Hashtbl.t = Hashtbl.create 100
 let name_def_prenorm : (string,trm) Hashtbl.t = Hashtbl.create 100
 let name_hyp : (string,trm) Hashtbl.t = Hashtbl.create 100
+let name_hyp_inv : (trm,string * trm) Hashtbl.t = Hashtbl.create 100 (*** associate a normalized/logic normalized term with the name of the assumption it came from and its original prenormalized form ***)
+let assumption_lit : (int,trm * trm) Hashtbl.t = Hashtbl.create 100 (*** associate assumption literals with their term after preprocessing and before preprocessing ***)
 
 (** Replaces all occurences of 'Neg' by 'implies False' **)
 let rec negnorm1 m =
@@ -89,8 +102,11 @@ let onlynegnorm m =
 let coqnorm m =
   let m = betanorm name_def_prenorm m in
   let (n,_) = negnorm1 m in n
+(** partially normalizes, without expanding certain defns (e.g., is_of) **)
+let partialnormalize pt = norm name_def (logicnorm pt)
 (** applies full satallax normalization**)
-let normalize pt = norm name_def (logicnorm pt)
+let normalize pt = norm name_def_all (logicnorm pt)
+let belnorm pt = norm name_def_empty (logicnorm pt)
 
 let coqknown (x,y) =
   if (!coq2) then y else x
@@ -155,6 +171,9 @@ let initial_branch_prenorm : trm list ref = ref []
 let processed : (trm,int) Hashtbl.t = Hashtbl.create 100
 let clause_ruleinfo : (clause,ruleinfo) Hashtbl.t ref = ref (Hashtbl.create 100)
 
+let allclauses : clause list ref = ref [] (* allclauses is only used if formdeps is activated, in which case someone wants useless information. it contains all the clauses that were used in all searches across different subgoals; doesn't get reset. Aug 2016 *)
+let allclause_ruleinfo : (clause,ruleinfo) Hashtbl.t ref = ref (Hashtbl.create 100)
+
 exception DuplicateClause
 
 let insert_clause c r =
@@ -166,10 +185,13 @@ let insert_clause c r =
   else
     begin
       Hashtbl.add clausesTable c ();
-      clauses := (c::!clauses);
+      clauses := c::!clauses;
+      if !mkproofterm = Some(PfFormdeps) then allclauses := c::!allclauses;
       match r with
       | Some(r) ->
 	  Hashtbl.add (!clause_ruleinfo) c r;
+	  if !mkproofterm = Some(PfFormdeps) then
+	    Hashtbl.add (!allclause_ruleinfo) c r;
       | None -> ()
     end
 
@@ -230,8 +252,9 @@ let print_proof_statistics minclauses =
   end
 
 
-let new_assumption_clause c =
+let new_assumption_lit l =
   try
+    let c = [l] in
     insert_clause c None;
     if not (add_clauses c) || not (minisat_search_period ()) then
       raise (Unsatisfiable None)
@@ -240,11 +263,10 @@ let new_assumption_clause c =
   | (Unsatisfiable _) ->
       if (!verbosity > 20) then Printf.printf "Proof found with assumption clauses only (no search)...%d clauses\n%!" (List.length (!clauses));
       if (mkprooftermp ()) then
-	raise (Unsatisfiable (Some AssumptionConflictR))
+	raise (Unsatisfiable (Some (AssumptionConflictR(l))))
       else
 	raise (Unsatisfiable None)
 
-
 let new_search_clause c r =
   try
     insert_clause c r;
@@ -267,8 +289,6 @@ let new_search_clause c r =
       else
 	raise (Unsatisfiable None)
 
-
-
 let patoms : (string,int * (trm list)) Hashtbl.t = Hashtbl.create 10
 let natoms : (string,int * (trm list)) Hashtbl.t = Hashtbl.create 10
 
@@ -357,7 +377,7 @@ let add_instantiation a m =
     add_instantiation_2 a m;
     set_default_elt_if_needed a m
 
-let choiceopnames : (string,(stp * trm)) Hashtbl.t = Hashtbl.create 10
+let choiceopnames : (string,(stp * trm * trm)) Hashtbl.t = Hashtbl.create 10
 
 let choiceop_axiom m =
   match m with
@@ -387,7 +407,7 @@ let choiceop_axiom m =
     -> Some(x,a)
   | _ -> None
 
-let declare_choiceop x a m = Hashtbl.add choiceopnames x (a,m)
+let declare_choiceop x a (m,mb) = Hashtbl.add choiceopnames x (a,m,mb)
 
 let choiceop m =
   match m with
@@ -395,7 +415,7 @@ let choiceop m =
   | Name(x,_) ->
       begin
 	try
-	  let (a,_) = Hashtbl.find choiceopnames x in
+	  let (a,_,_) = Hashtbl.find choiceopnames x in
 	  Some(a)
 	with
 	| Not_found -> None
@@ -611,7 +631,6 @@ let print_coqsig c =
                 (*have Syntax.trm but need Syntax.pretrm, so look it up*)
               let pt = List.assoc x !coqsig_hyp in
 	      begin
-	        print_coqsig_hyp r;
 	        Printf.fprintf c "Hypothesis %s : " x;
 	        if (!coq2) then print_pretrm_coq2 c pt (-1) (-1) else print_pretrm_coq c pt coq_names coq_used_names (-1) (-1);
 	        Printf.fprintf c ".\n"
@@ -658,7 +677,7 @@ let print_coqsig c =
 	  Printf.fprintf c "\nlemma\n";
         print_coqsig_hyp !coqsig_hyp_trm;
         match (!conjecture) with
-          | Some (m,t,_) ->
+          | Some (t,_) ->
               if !mkproofterm = Some IsarScript then
 	        begin
 	          Printf.fprintf c "shows %s : \"" (Hashtbl.find coq_hyp_names (!conjecturename));
@@ -675,7 +694,7 @@ let print_coqsig c =
               else
 	        begin
 	          Printf.fprintf c "Theorem %s : " (Hashtbl.find coq_hyp_names (!conjecturename));
-	          if (!coq2) then print_pretrm_coq2 c m (-1) (-1) else print_pretrm_coq c m coq_names coq_used_names (-1) (-1);
+(*	          if (!coq2) then print_pretrm_coq2 c m (-1) (-1) else print_pretrm_coq c m coq_names coq_used_names (-1) (-1); *) (*** No longer supported to get rid of dependence on pretrm ***)
 	          Printf.fprintf c ".\n"
 	        end
           | None ->
@@ -732,9 +751,36 @@ let st_to_trm m =
 	end
   end
 
-let declare_typed_constant (name:string) (role:string) (m:pretrm) =
+let is_of_names : string list ref = ref []
+let all_of_names : string list ref = ref []
+let is_of_name : (string,unit) Hashtbl.t = Hashtbl.create 1
+let all_of_name : (string,unit) Hashtbl.t = Hashtbl.create 1
+
+let is_of_p m =
+  match m with
+  | Name(x,Ar(Base("$i"),Ar(Ar(Base("$i"),Prop),Prop))) -> Hashtbl.mem is_of_name x
+  | Lam(Base("$i"),Lam(Ar(Base("$i"),Prop),Ap(DB(0,Ar(Base("$i"),Prop)),DB(1,Base("$i"))))) -> true
+  | _ -> false
+	
+let all_of_p m =
+  match m with
+  | Name(x,Ar(Ar(Base("$i"),Prop),Ar(Ar(Base("$i"),Prop),Prop))) -> Hashtbl.mem all_of_name x
+  | Lam(Ar(Base("$i"),Prop),Lam(Ar(Base("$i"),Prop),Ap(Forall(Base("$i")),Lam(Ar(Base("$i"),Prop),Ap(Ap(Imp,Ap(Ap(m1,DB(0,Base("$i"))),DB(2,Ar(Base("$i"),Prop)))),DB(1,Ar(Base("$i"),Prop))))))) ->
+      is_of_p m1
+  | _ -> false
+
+let rec ontology_prop_p m =
+  match m with
+  | Ap(Ap(Name(x,Ar(Ar(Base("$i"),Prop),Ar(Ar(Base("$i"),Prop),Prop))),m1),m2) when Hashtbl.mem all_of_name x -> ontology_prop_p (gen_lam_body (Base("$i")) m2)
+  | Ap(Ap(Name(x,Ar(Base("$i"),Ar(Ar(Base("$i"),Prop),Prop))),m1),m2) when Hashtbl.mem is_of_name x -> true
+  | Ap(Forall(a),m1) -> ontology_prop_p (gen_lam_body a m1)
+  | Ap(Ap(Imp,m1),m2) -> ontology_prop_p m2
+  | _ -> false
+
+let declare_typed_constant (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
   begin
     if (!verbosity > 4) then (Printf.printf "declare_typed_constant %s %s\n%s\n" name role (pretrm_str m); flush stdout);
+    if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
     match m with
       POf(PName(x),m) ->
 	begin
@@ -748,7 +794,9 @@ let declare_typed_constant (name:string) (role:string) (m:pretrm) =
 	      if (Hashtbl.mem name_tp x) then raise (Redeclaration x);
 	      declare_base_type x
 	  | _ ->
-	      let tp = (st_to_stp m) in
+	      let tp = st_to_stp m in
+	      if !recognize_hashroots && tp = Ar(Base("$i"),Ar(Ar(Base("$i"),Prop),Prop)) && List.mem ("hashroot","01b8315391fd465f550f1a3788664028f3a6126b6bded775854e3b7d33765cd9") al then (is_of_names := x::!is_of_names; Hashtbl.add is_of_name x ()); (*** this annotation signals the prover to treat the name is the ontological "of" relation ***)
+	      if !recognize_hashroots && tp = Ar(Ar(Base("$i"),Prop),Ar(Ar(Base("$i"),Prop),Prop)) && List.mem ("hashroot","67f26991e531d6c57a2161d295d3cdf1f416713aa711d23ccb703b2431633300") al then (all_of_names := x::!all_of_names; Hashtbl.add all_of_name x ()); (*** this annotation signals the prover to treat the name is the ontological "all_of" quantifier ***)
 	      if (Hashtbl.mem name_base x) then raise (Redeclaration x);
 	      if (Hashtbl.mem name_tp x) then raise (Redeclaration x);
 	      if (!coq) then
@@ -762,7 +810,61 @@ let declare_typed_constant (name:string) (role:string) (m:pretrm) =
 	end
     | _ -> raise (GenericError ("Incorrect format for type declaration " ^ name))
   end
-    
+
+let translucent_defn_p m =
+  get_bool_flag "ONTOLOGY_DEFS_TRANSLUCENT" && is_of_p m
+
+let rec init_probitem m =
+  match m with
+  | ProbDef(x,a,m,al,w) ->
+      begin
+	try
+	  if (get_bool_flag "ALL_DEFS_AS_EQNS") then
+	    raise TreatAsAssumption
+	  else
+	    let m0 = logicnorm m in
+	    let m2 = norm name_def m0 in
+	    if (!verbosity > 20) then Printf.printf "name_def %s %s\n" x (trm_str m2);
+	    Hashtbl.add name_def_all x (norm name_def_all m2);
+	    if not (translucent_defn_p m) then
+	      begin
+		translucent_defns := true;
+		Hashtbl.add name_def x m2
+	      end
+	with TreatAsAssumption ->
+	  init_probitem (ProbAx(x,"hypothesis",Ap(Ap(Eq(a),Name(x,a)),m),al,w))
+      end
+  | ProbAx(name,role,tm,al,w) ->
+      begin
+	let m0 = logicnorm tm in
+	let tmn = norm name_def m0 in
+	Hashtbl.add name_hyp name tmn;
+	Hashtbl.add name_hyp_inv tmn (name,tm);
+	if (!verbosity > 20) then Printf.printf "name_hyp %s %s\n" name (trm_str tmn);
+	initial_branch_prenorm := tm::(!initial_branch_prenorm);
+	initial_branch := tmn::(!initial_branch)
+      end
+  | ProbConj(name,tm,al,w) ->
+      begin
+	conjecturename := name;
+	if (name != "claim") then
+	  ignore (coqify_name name coq_hyp_names coq_used_names);
+	let ntm = Ap(Neg,tm) in
+	let ntmn = norm name_def (logicnorm ntm) in
+	initial_branch_prenorm := (ntm::(!initial_branch_prenorm));
+	if (!verbosity > 20) then Printf.printf "name_conj negated %s %s\n" name (trm_str ntmn);
+	initial_branch := (ntmn::(!initial_branch));
+	conjecture := Some (tm,ntmn);
+	Hashtbl.add part_of_conjecture ntmn ()
+      end
+
+let defpreweight m al = 2.0
+
+let axpreweight m al =
+  if ontology_prop_p m then 1.0 else 3.0
+
+let conjpreweight m al = 0.0
+
 (***
  THF official policy requires that all names are given a type before use,
  that definitions can be circular (in which case they must be treated as equations),
@@ -777,13 +879,13 @@ let declare_typed_constant (name:string) (role:string) (m:pretrm) =
  NOTE: If I wanted to be strict, I would exit if the definition is given without the name being given a type,
  but I will not require the type to be given.  I will print a warning if the verbosity is > 0.
  ***)
-let declare_definition_real x m =
+let declare_definition_real x m al =
   if (Hashtbl.mem name_base x) then raise (Redeclaration x);
   if (Hashtbl.mem name_def x) then raise TreatAsAssumption; (*** treat it as an assumption per THF policy. - Mar 31, 2011 - Chad ***)
 (*** raise (Redeclaration x); ***)
   try
     let tp = Hashtbl.find name_tp x in
-    let tm = st_to_trm_given_stp m tp in
+    let tm = belnorm (st_to_trm_given_stp m tp) in
     try 
       let (_,r) = Hashtbl.find name_trm x in
       if (!r) then
@@ -793,7 +895,7 @@ let declare_definition_real x m =
     with
     | Not_found ->
 	begin
-	  if (!coq) then
+	  if (!coq) then (*** Coq proofs now only work for non-slaves ***)
 	    begin (*** The name was coqified when the type was declared. ***)
 	      try
 		let y = Hashtbl.find coq_names x in
@@ -803,13 +905,19 @@ let declare_definition_real x m =
 	      | Not_found -> raise (GenericError("Could not find Coq version of name " ^ x))
 	    end;
 	  Hashtbl.add name_def_prenorm x tm;
-	  Hashtbl.add name_def x (norm name_def (logicnorm tm))
+	  if is_of_p tm then
+	    (is_of_names := x::!is_of_names; Hashtbl.add is_of_name x ())
+	  else if all_of_p tm then
+	    (all_of_names := x::!is_of_names; Hashtbl.add all_of_name x ());
+	  let w = defpreweight tm al in
+	  probsig := ProbDef(x,tp,tm,al,w)::!probsig
 	end
   with
   | Not_found ->
       begin (*** Giving a definition without giving it's type isn't allowed in THF anymore.  I'm still allowing it.  ***)
 	if ((!verbosity > 0) && (not (!coqglobalfile)) && (not (!coqlocalfile))) then Printf.printf "WARNING: %s defined without giving type first.\n" x;
 	let (tm,tp) = st_to_trm m in
+	let tm = belnorm tm in
 	if (!coq) then
 	  begin
 	    let y = coqify_name x coq_names coq_used_names in
@@ -821,66 +929,59 @@ let declare_definition_real x m =
 	Hashtbl.add name_trm x ((Name(x,tp),tp),ref false);
 	name_trm_list := (x,Name(x,tp),tp)::!name_trm_list;
 	Hashtbl.add name_def_prenorm x tm;
-	Hashtbl.add name_def x (norm name_def (logicnorm tm))
+	let w = defpreweight tm al in
+	probsig := ProbDef(x,tp,tm,al,w)::!probsig
       end
 
-let rec declare_thf_logic_formula (name:string) (role:string) (m:pretrm) =
+let rec declare_thf_logic_formula (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
   begin
     if !verbosity > 4 then (Printf.printf "declare_thf_logic_formula %s %s\n" name role; flush stdout);
-    if ((role = "axiom") || (role = "hypothesis") || (role = "lemma")) then
+    if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
+    if ((role = "axiom") || (role = "hypothesis") || (role = "assumption") || (role = "lemma") || (role = "theorem") || (role = "corollary")) then
       begin
 	if (Hashtbl.mem name_hyp name) then raise (Redeclaration name);
 	let tm = st_to_trm_given_stp m Prop in
-	if (!coq) then
+	let tm = belnorm tm in
+	if (!coq) then (** Giving Coq proofs will now only work for non-slaves since the pretrm is used **)
 	  begin
 	    let y = coqify_name name coq_hyp_names coq_used_names in
 	    coqsig_hyp := ((y,m)::!coqsig_hyp);
 	    coqsig_hyp_trm := ((y,tm)::!coqsig_hyp_trm)
 	  end;
-	let tmn = norm name_def (logicnorm tm) in
-	Hashtbl.add name_hyp name tmn;
-	if (!verbosity > 20) then Printf.printf "name_hyp %s %s\n" name (trm_str tmn);
-	initial_branch_prenorm := (tm::(!initial_branch_prenorm));
-	initial_branch := (tmn::(!initial_branch))
+	let w = axpreweight tm al in
+	probsig := ProbAx(name,role,tm,al,w)::!probsig
       end
-    else if ((role = "conjecture") || (role = "theorem")) then
+    else if (role = "conjecture") then
       begin
 	match (!conjecture) with
 	| Some _ -> raise (GenericError "Problem file has more than one conjecture.")
 	| None ->
 	    let tm = st_to_trm_given_stp m Prop in
-	    begin
-	      conjecturename := name;
-	      if (name != "claim") then
-		ignore (coqify_name name coq_hyp_names coq_used_names);
-	      let ntm = Ap(Neg,tm) in
-	      let ntmn = norm name_def (logicnorm ntm) in
-	      initial_branch_prenorm := (ntm::(!initial_branch_prenorm));
-	      initial_branch := (ntmn::(!initial_branch));
-	      conjecture := Some (m,tm,ntmn);
-	      Hashtbl.add part_of_conjecture ntmn ()
-	    end
+	    let tm = belnorm tm in
+	    let w = conjpreweight tm al in
+	    probsig := ProbConj(name,tm,al,w)::!probsig
       end
     else
       raise (GenericError ("Unknown role " ^ role))
   end
-and declare_definition (name:string) (role:string) (m:pretrm) =
+and declare_definition (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
   try
     begin
       if (get_bool_flag "ALL_DEFS_AS_EQNS") then raise TreatAsAssumption;
       if !verbosity > 4 then (Printf.printf "declare_definition %s %s\n" name role; flush stdout);
+      if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
       match m with
 	PDef(PName(x),m) -> (*** No longer THF syntax. ***)
-	  declare_definition_real x m
+	  declare_definition_real x m al
       | PAp(PAp(PEq,PName(x)),m) ->
-	  declare_definition_real x m
-      | _ -> (*** Treat as an assumption, now matter how it looks. Odd, but OK. This may be too liberal; we haven't decided yet. ***)
+	  declare_definition_real x m al
+      | _ -> (*** Treat as an assumption, no matter how it looks. Odd, but OK. This may be too liberal; we haven't decided yet. ***)
 	  raise TreatAsAssumption
 (*** raise (GenericError ("Incorrect format for definition " ^ name)) ***)
     end
   with
   | TreatAsAssumption ->
-      declare_thf_logic_formula name "hypothesis" m
+      declare_thf_logic_formula name "hypothesis" m al
 
 (*** Code for enumeration of types and terms - Dec 10, 2010 - Chad ***)
 let enum_started = ref false
