@@ -16,6 +16,7 @@ val num_axioms : unit -> int
 val slaveargs : string list ref
 val mode : string list ref
 val timeout : float option ref
+val hardtimeout : float option ref
 val nontheorem : bool ref
 val coq : bool ref
 val coq2 : bool ref
@@ -30,13 +31,28 @@ val coqctx : prectx ref
 val coqglobalsectionstack : (string * (out_channel -> unit) * prectx) list ref
 val updatecoqglobalsectionstack : prectx -> (string * (out_channel -> unit) * prectx) list -> (prectx -> (out_channel -> unit) -> (out_channel -> unit)) -> (string * (out_channel -> unit) * prectx) list
 
+val is_of_names : string list ref
+val all_of_names : string list ref
+val is_of_name : (string,unit) Hashtbl.t
+val all_of_name : (string,unit) Hashtbl.t
+
+type probitem =
+  | ProbDef of string * stp * trm * (string * string) list * float
+  | ProbAx of string * string * trm * (string * string) list * float
+  | ProbConj of string * trm * (string * string) list * float
+val probsig : probitem list ref
+val init_probitem : probitem -> unit
+val recognize_hashroots : bool ref
+
 val training_file : string option ref
 
 val conjecturename : string ref
-val conjecture : (pretrm * trm * trm) option ref
-type proofkind = TSTP | CoqScript | CoqSPfTerm | HOCore | Model | ModelTrue | IsarScript
+val conjecture : (trm * trm) option ref
+type proofkind = TSTP | CoqScript | CoqSPfTerm | HOCore | Model | ModelTrue | IsarScript | PfInfo | PfUseful | PfFormdeps
 val mkproofterm : proofkind option ref
 val mkprooftermp : unit -> bool
+val pfusefulout : string option ref
+val pfformdepsout : string option ref
 val slave : bool ref
 val coqsig_base : string list ref
 val coqsig_const : (string * stp) list ref
@@ -49,11 +65,16 @@ val name_base_list : string list ref
 val name_tp : (string,stp) Hashtbl.t
 val name_trm : (string,(trm * stp) * bool ref) Hashtbl.t
 val name_trm_list : (string * trm * stp) list ref
+val translucent_defns : bool ref
 val name_def : (string,trm) Hashtbl.t
+val name_def_all : (string,trm) Hashtbl.t
 val name_def_prenorm : (string,trm) Hashtbl.t
 val name_hyp : (string,trm) Hashtbl.t
+val name_hyp_inv : (trm,string * trm) Hashtbl.t
+val assumption_lit : (int,trm * trm) Hashtbl.t
 val coqknown : string * string -> string
-
+val completep : bool ref
+    
 val mult_timeout : float -> unit
 
 val required : string ref
@@ -69,9 +90,12 @@ val clauses : clause list ref
 val processed : (trm,int) Hashtbl.t
 val clause_ruleinfo : (clause,ruleinfo) Hashtbl.t ref
 
+val allclauses : clause list ref (* allclauses is only used if formdeps is activated, in which case someone wants useless information. it contains all the clauses that were used in all searches across different subgoals. Aug 2016 *)
+val allclause_ruleinfo : (clause,ruleinfo) Hashtbl.t ref
+
 exception DuplicateClause
 
-val new_assumption_clause : clause -> unit
+val new_assumption_lit : int -> unit
 val new_search_clause : clause -> ruleinfo option -> unit
 
 
@@ -102,13 +126,13 @@ val known_instantiation : stp -> trm -> bool
 val add_instantiation : stp -> trm -> unit
 
 (*** Hash table associating names of epsilon operators with (a,m) where a is the type and m is the formula giving the choice axiom ***)
-val choiceopnames : (string,(stp * trm)) Hashtbl.t
+val choiceopnames : (string,(stp * trm * trm)) Hashtbl.t
 
 (*** Check if a formula says that a name is a choice operator at some type ***)
 val choiceop_axiom : trm -> (string * stp) option
 
 (*** Declare a name to be a choice operator at some type ***)
-val declare_choiceop : string -> stp -> trm -> unit
+val declare_choiceop : string -> stp -> trm * trm -> unit
 
 (*** If trm is a choice operator at some type, return the type, otherwise None ***)
 val choiceop : trm -> stp option
@@ -135,11 +159,13 @@ val st_find_read_thf_fun : (string -> string -> unit) ref
 val coq_init : unit -> unit
 val print_coqsig : out_channel -> unit
 
-val declare_typed_constant : string -> string -> pretrm -> unit
-    
-val declare_definition : string -> string -> pretrm -> unit
+val declare_base_type : string -> unit
 
-val declare_thf_logic_formula : string -> string -> pretrm -> unit
+val declare_typed_constant : string -> string -> pretrm -> (string * string) list -> unit
+    
+val declare_definition : string -> string -> pretrm -> (string * string) list -> unit
+
+val declare_thf_logic_formula : string -> string -> pretrm -> (string * string) list -> unit
 
 (*** Code for enumeration of types and terms ***)
 val enum_started : bool ref
