@@ -69,6 +69,9 @@ val choice:stp -> trm -> trm
 
 val stp_str:stp -> string
 val trm_str:trm -> string
+val next_var_name : stp -> string list -> string list -> string
+val trm_str_nice_rec:trm -> string list -> string list -> string * int
+val trm_str_nice:trm -> string (*** gives a version with named vars and ~, ->, ! intended to be more readable for the humans ***)
 
 val tpof:trm -> stp
 
